//******************************************************************************
//! @file
//! @brief Event template definitions.
//******************************************************************************

#ifndef UTIL_EVENT_H
#define UTIL_EVENT_H

#include "fun_ptr3.h"
#include "finally.h"
#include <vector>
#include <algorithm>
#include <stdexcept>
#include <iterator>
#include <cassert>

namespace util
{

template <typename FuncSignature>
class event;

//! @brief event class
//! @tparam Ret The return type of the event function signature (usually void).
//! @tparam Args event function arguments
template <typename Ret, typename... Args>
class event<Ret(Args...)>
{
public:
   event(const event&) = delete;
   event(event&&) = delete;
   event& operator=(const event&) = delete;
   event& operator=(event&&) = delete;

   //! @brief Define the fun_ptr type used to call functions with
   //!  the function signature of this event.
   typedef util3::fun_ptr<Ret(Args...)> delegate_type;

   //! @brief Defines the type used to invoke this event.
   typedef delegate_type invoker_type;

   event() : delegate_list_(), p_invoke_data_{nullptr}, gave_invoker_{false}
   {
   }

   //! @brief retrieve the invoker for this event.
   //!
   //! This function can only be called once.
   invoker_type get_invoker();

   //! @brief Add a delegate function to the event.
   void attach(const delegate_type& theDelegate)
   {
      delegate_list_.push_back(theDelegate);
   }

   //! @brief Remove a delegate function from the event.
   void detach(const delegate_type& theDelegate);

private:
   // Calls all the attached delegates. The invoker function object calls this
   // function.
   Ret invoke(Args... args);

   // a signed type is needed for the index type because a value if -1 might
   // exist.
   typedef typename std::vector<delegate_type>::difference_type index_type;

   struct invoke_data {
      index_type current; // index to the current delegate
      index_type end;     // index to one past the last delegate
   };

   std::vector<delegate_type> delegate_list_;
   invoke_data* p_invoke_data_;
   bool gave_invoker_;
};

template <typename Ret, typename... Args>
typename event<Ret(Args...)>::invoker_type event<Ret(Args...)>::get_invoker()
{
   if (gave_invoker_) {
      throw ::std::runtime_error("cannot give multiple event invokers");
   }
   gave_invoker_ = true;
   return util3::make_fun(&event<Ret(Args...)>::invoke, this);
}

template <typename Ret, typename... Args>
void event<Ret(Args...)>::detach(const delegate_type& theDelegate)
{
   // Find the delegate to erase (use reverse iterator for last in first out
   // semantics).
   auto erase_rit =
      std::find(delegate_list_.rbegin(), delegate_list_.rend(), theDelegate);

   // Was a match found?
   if (erase_rit != delegate_list_.rend()) {
      // match found - convert the reverse iterator to an iterator
      auto erase_it = (++erase_rit).base();

      // Is this event currently in the context of invoke()?
      if (p_invoke_data_) {
         // Care must be taken if detaching from the context of the invoke()
         // function since the vector is currently being iterated over. Invoke
         // data may need to be updated.

         // convert eraseIt to an index (guaranteed to be >= 0)
         index_type erase_index = std::distance(delegate_list_.begin(), erase_it);

         // If the element being erased is < than the currently being invoked
         // delegate (p_invoke_data_->current) the index needs to be adjusted left
         // one slot so the index points to the same element. If the element
         // being erased is what current points to (the == case) the index still
         // needs to be adjusted left one since upon returning to invoke() the
         // index is advanced. note: current may be decremented to -1 which is
         // well defined whereas delegate_list_.begin() - 1 is undefined land
         // which is one reason why an index is being used instead of an
         // iterator.
         if (erase_index <= p_invoke_data_->current) {
            assert(p_invoke_data_->current < p_invoke_data_->end);
            --p_invoke_data_->current;
            --p_invoke_data_->end;
         } else if (erase_index < p_invoke_data_->end) {
            --p_invoke_data_->end;
         }
      }

      // safely erase element
      delegate_list_.erase(erase_it);
   }
}

template <typename Ret, typename... Args>
Ret event<Ret(Args...)>::invoke(Args... args)
{
   // ensure that none of the subscribed delegates invoke the event.
   assert(!p_invoke_data_);

   auto invoke_data_scoped =
      invoke_data{0, static_cast<index_type>(delegate_list_.size())};

   // begin invoke context
   p_invoke_data_ = &invoke_data_scoped;
   auto invokeContext = util::finally([&] {
      // end of invoke context
      p_invoke_data_ = nullptr;
   });

   // call each delegate
   while (invoke_data_scoped.current < invoke_data_scoped.end) {
      // note: the call may alter current and end if callback calls += or -=
      delegate_list_[invoke_data_scoped.current](args...);
      assert(invoke_data_scoped.current >= -1);
      assert(invoke_data_scoped.current < invoke_data_scoped.end);
      ++invoke_data_scoped.current;
   }
}

} // namespace util

#endif // UTIL_EVENT_H
