#ifndef FUN_PTR2_H
#define FUN_PTR2_H

#include <memory>  //std::shared_ptr
#include <utility> //std::forward

namespace util2
{

namespace detail
{

template <typename Ret, typename... Args>
class target_base
{
public:
   virtual ~target_base() = default;
   virtual Ret invoke(Args... args) const = 0;
   virtual bool check_equality(const target_base& other) const noexcept = 0;

protected:
   target_base() noexcept = default;
};

template <typename Ret, typename... Args>
class standalone_target : public target_base<Ret, Args...>
{
public:
   explicit standalone_target(Ret (*func)(Args...)) noexcept
      : target_base<Ret, Args...>(),
        func_(func)
   {
   }

private:
   Ret invoke(Args... args) const override
   {
      return (*func_)(std::forward<Args>(args)...);
   }

   bool check_equality(const target_base<Ret, Args...>& other) const
      noexcept override
   {
      auto p = dynamic_cast<const standalone_target*>(&other);
      return (p != nullptr) && (func_ == p->func_);
   }

   Ret (*const func_)(Args...);
};

template <typename Ret, typename T, typename... Args>
class mutable_member_target : public target_base<Ret, Args...>
{
public:
   explicit mutable_member_target(Ret (T::*method)(Args...), T* object) noexcept
      : target_base<Ret, Args...>(),
        method_(method),
        object_(object)
   {
   }

private:
   Ret invoke(Args... args) const override
   {
      return (object_->*method_)(args...);
   }

   bool check_equality(const target_base<Ret, Args...>& other) const
      noexcept override
   {
      auto p = dynamic_cast<const mutable_member_target*>(&other);
      return (p != nullptr) && (object_ == p->object_) &&
             (method_ == p->method_);
   }

   Ret (T::*const method_)(Args...);
   T* const object_;
};

template <typename Ret, typename T, typename... Args>
class const_member_target : public target_base<Ret, Args...>
{
public:
   explicit const_member_target(Ret (T::*method)(Args...) const,
                                const T* object) noexcept
      : target_base<Ret, Args...>(),
        method_(method),
        object_(object)
   {
   }

private:
   Ret invoke(Args... args) const override
   {
      return (object_->*method_)(args...);
   }

   bool check_equality(const target_base<Ret, Args...>& other) const
      noexcept override
   {
      auto p = dynamic_cast<const const_member_target*>(&other);
      return (p != nullptr) && (object_ == p->object_) &&
             (method_ == p->method_);
   }

   Ret (T::*const method_)(Args...) const;
   const T* const object_;
};

template <typename Ret, typename T, typename... Args>
class function_object_target : public target_base<Ret, Args...>
{
public:
   explicit function_object_target(T f) noexcept : target_base<Ret, Args...>(),
                                                   f_(f)
   {
   }

private:
   Ret invoke(Args... args) const override
   {
      return f_(args...);
   }

   bool check_equality(const target_base<Ret, Args...>& other) const
      noexcept override
   {
      auto p = dynamic_cast<const function_object_target*>(&other);
      return (p != nullptr) && (&f_ == &p->f_);
   }

   mutable T f_;
};

} // namespace detail

template <typename FuncSignature>
class fun_ptr;

template <typename Ret, typename... Args>
class fun_ptr<Ret(Args...)>
{
public:
   fun_ptr() noexcept : fun_ptr{nullptr}
   {
   }

   fun_ptr(std::nullptr_t) noexcept : target_(nullptr)
   {
   }

   fun_ptr(Ret (*func)(Args...))
      : target_(std::make_shared<detail::standalone_target<Ret, Args...>>(func))
   {
   }

   template <typename T>
   fun_ptr(Ret (T::*method)(Args...), T* object)
      : target_(
           std::make_shared<detail::mutable_member_target<Ret, T, Args...>>(
              method, object))
   {
   }

   template <typename T>
   fun_ptr(Ret (T::*method)(Args...) const, const T* object)
      : target_(std::make_shared<detail::const_member_target<Ret, T, Args...>>(
           method, object))
   {
   }

   template <typename T>
   fun_ptr(T f)
      : target_(
           std::make_shared<detail::function_object_target<Ret, T, Args...>>(f))
   {
   }

   fun_ptr(const fun_ptr&) noexcept = default;

   fun_ptr(fun_ptr&&) noexcept = default;

   ~fun_ptr() noexcept = default;

   fun_ptr& operator=(std::nullptr_t) noexcept
   {
      target_ = nullptr;
      return *this;
   }

   fun_ptr& operator=(const fun_ptr&) noexcept = default;

   fun_ptr& operator=(fun_ptr&&) noexcept = default;

   explicit operator bool() const noexcept
   {
      return target_ != nullptr;
   }

   template <typename... FwArgs>
   Ret operator()(FwArgs&&... args) const
   {
      // note: hard undefined behavior if target_ is nullptr
      return target_->invoke(std::forward<FwArgs>(args)...);
   }

private:
   std::shared_ptr<detail::target_base<Ret, Args...>> target_;

   template <typename FrRet, typename... FrArgs>
   friend bool operator==(const fun_ptr<FrRet(FrArgs...)>& lhs,
                          const fun_ptr<FrRet(FrArgs...)>& rhs) noexcept;
};

template <typename Ret, typename... Args>
inline bool operator==(const fun_ptr<Ret(Args...)>& lhs,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept
{
   if (lhs.target_) {
      if (rhs.target_) {
         // both not nullptr
         return lhs.target_->check_equality(*rhs.target_);
      }
      return false;
   }

   if (rhs.target_) {
      return false;
   }

   // both nullptr
   return true;
}

template <typename Ret, typename... Args>
inline bool operator!=(const fun_ptr<Ret(Args...)>& lhs,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept
{
   return !(lhs == rhs);
}

template <typename Ret, typename... Args>
inline bool operator==(const fun_ptr<Ret(Args...)>& lhs,
                       std::nullptr_t) noexcept
{
   return !static_cast<bool>(lhs);
}

template <typename Ret, typename... Args>
inline bool operator!=(const fun_ptr<Ret(Args...)>& lhs,
                       std::nullptr_t) noexcept
{
   return !(lhs == nullptr);
}

template <typename Ret, typename... Args>
inline bool operator!=(std::nullptr_t,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept
{
   return !(rhs == nullptr);
}

template <typename Ret, typename... Args>
inline bool operator==(std::nullptr_t,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept
{
   return rhs == nullptr;
}

template <typename Ret, typename... Args>
inline fun_ptr<Ret(Args...)> make_fun(Ret (*fp)(Args...))
{
   return fun_ptr<Ret(Args...)>(fp);
}

template <typename Ret, typename T, typename... Args>
inline fun_ptr<Ret(Args...)> make_fun(Ret (T::*fp)(Args...), T* obj)
{
   return fun_ptr<Ret(Args...)>(fp, obj);
}

template <typename Ret, typename T, typename... Args>
inline fun_ptr<Ret(Args...)> make_fun(Ret (T::*fp)(Args...) const, const T* obj)
{
   return fun_ptr<Ret(Args...)>(fp, obj);
}

template <typename T>
inline auto make_fun(T lambda)
   -> decltype(make_fun(&T::operator(), static_cast<T*>(nullptr)))
{
   return decltype(make_fun(&T::operator(), static_cast<T*>(nullptr)))(lambda);
}

} // // namespace util2

#endif // FUN_PTR2_H
