#ifndef UTIL_FUN_PTR3_H
#define UTIL_FUN_PTR3_H

#include <new>         // placement new
#include <utility>     // std::forward
#include <type_traits> // std::aligned_union
#include <memory>      // std::shared_ptr
#include <stdexcept>   // std::runtime_error
#include <cassert>     // assert
#if !defined(NDEBUG)
#include <typeinfo>
#endif
#include "aligned_union.h" // until gcc has one
#include <iostream>

namespace util3
{

namespace detail
{

// type used to uniquely identify a list of template types
typedef void (*unique_id_t)();

// each instantiation of this template has its own static dummy function. The
// address of this function is used to uniquely identify the list of types
template <typename... Args>
struct IdGen {
   static constexpr inline unique_id_t get_unique_id()
   {
      return &IdGen::dummy;
   }

private:
   static void dummy()
   {
   }
};

template <typename Ret, typename... Args>
class target_base
{
public:
   virtual ~target_base() noexcept = default;

   virtual Ret invoke(Args... args) const = 0;

   // virtual constructor idiom - deep copy the object
   virtual target_base* clone(void* ptr) const noexcept = 0;

   virtual bool check_equality(const target_base& other) const noexcept = 0;

protected:
   explicit target_base() noexcept = default;
};

template <typename Ret, typename... Args>
class null_target : public target_base<Ret, Args...>
{
public:
   explicit null_target() noexcept : target_base<Ret, Args...>()
   {
   }

private:
   Ret invoke(Args... /*args*/) const override
   {
      assert(!"invoked a null_target");
      throw std::runtime_error("invoked a null_target");
   }

   bool check_equality(const target_base<Ret, Args...>& other) const
      noexcept override
   {
      assert(typeid(*this) == typeid(other));
      (void)other; // quiet unused variable warning when debug is disabled
      return true;
   }

   target_base<Ret, Args...>* clone(void* ptr) const noexcept override
   {
      // since this object fit into the size ptr has the clone will also fit
      return new (ptr) null_target();
   }
};

template <typename Ret, typename... Args>
class standalone_target : public target_base<Ret, Args...>
{
public:
   explicit standalone_target(Ret (*func)(Args...)) noexcept
      : target_base<Ret, Args...>(),
        func_(func)
   {
   }

private:
   Ret invoke(Args... args) const override
   {
      return (*func_)(args...);
   }

   bool check_equality(const target_base<Ret, Args...>& other) const
      noexcept override
   {
      assert(typeid(*this) == typeid(other));
      auto& r = static_cast<const standalone_target&>(other);
      return (func_ == r.func_);
   }

   target_base<Ret, Args...>* clone(void* ptr) const noexcept override
   {
      // since this object fit into the size ptr has the clone will also fit
      return new (ptr) standalone_target(func_);
   }

   Ret (*const func_)(Args...);
};

template <typename Ret, typename T, typename... Args>
class mutable_member_target : public target_base<Ret, Args...>
{
public:
   explicit mutable_member_target(Ret (T::*method)(Args...), T* object) noexcept
      : target_base<Ret, Args...>(),
        method_(method),
        object_(object)
   {
   }

private:
   Ret invoke(Args... args) const override
   {
      return (object_->*method_)(args...);
   }

   bool check_equality(const target_base<Ret, Args...>& other) const
      noexcept override
   {
      assert(typeid(*this) == typeid(other));
      auto& r = static_cast<const mutable_member_target&>(other);
      return (method_ == r.method_) && (object_ == r.object_);
   }

   target_base<Ret, Args...>* clone(void* ptr) const noexcept override
   {
      // since this object fit into the size ptr has the clone will also fit
      return new (ptr) mutable_member_target(method_, object_);
   }

   Ret (T::*const method_)(Args...);
   T* const object_;
};

template <typename Ret, typename T, typename... Args>
class const_member_target : public target_base<Ret, Args...>
{
public:
   explicit const_member_target(Ret (T::*method)(Args...) const,
                                const T* object) noexcept
      : target_base<Ret, Args...>(),
        method_(method),
        object_(object)
   {
   }

private:
   Ret invoke(Args... args) const override
   {
      return (object_->*method_)(args...);
   }

   bool check_equality(const target_base<Ret, Args...>& other) const
      noexcept override
   {
      assert(typeid(*this) == typeid(other));
      auto& r = static_cast<const const_member_target&>(other);
      return (method_ == r.method_) && (object_ == r.object_);
   }

   target_base<Ret, Args...>* clone(void* ptr) const noexcept override
   {
      return new (ptr) const_member_target(method_, object_);
   }

   Ret (T::*const method_)(Args...) const;
   const T* const object_;
};

template <typename Ret, typename T, typename... Args>
class function_object_target : public target_base<Ret, Args...>
{
public:
   explicit function_object_target(T f)
      : target_base<Ret, Args...>(), pf_(std::make_shared<T>(f))
   {
   }

private:
   explicit function_object_target(std::shared_ptr<T> pf) noexcept
      : target_base<Ret, Args...>(),
        pf_(std::move(pf))
   {
   }

   Ret invoke(Args... args) const override
   {
      return (*pf_)(args...);
   }

   bool check_equality(const target_base<Ret, Args...>& other) const
      noexcept override
   {
      assert(typeid(*this) == typeid(other));
      auto& r = static_cast<const function_object_target&>(other);
      return pf_ == r.pf_;
   }

   target_base<Ret, Args...>* clone(void* ptr) const noexcept override
   {
      return new (ptr) function_object_target(pf_);
   }

   std::shared_ptr<T> pf_;
};

class SomeClass;

} // namespace detail

template <typename FuncSignature>
class fun_ptr;

template <typename Ret, typename... Args>
class fun_ptr<Ret(Args...)>
{
public:
   fun_ptr() noexcept : fun_ptr{nullptr}
   {
   }

   fun_ptr(std::nullptr_t) noexcept
      : mem_type_id_{
           detail::IdGen<detail::null_target<Ret, Args...>>::get_unique_id()}
   {
      static_assert(alignof(decltype(mem_)) %
                          alignof(detail::null_target<Ret, Args...>) ==
                       0,
                    "");
      static_assert(sizeof(mem_) >= sizeof(detail::null_target<Ret, Args...>),
                    "");
      new (&mem_) detail::null_target<Ret, Args...>();
   }

   fun_ptr(Ret (*func)(Args...))
      : mem_type_id_{detail::IdGen<
           detail::standalone_target<Ret, Args...>>::get_unique_id()}
   {
      static_assert(alignof(decltype(mem_)) %
                          alignof(detail::standalone_target<Ret, Args...>) ==
                       0,
                    "");
      static_assert(
         sizeof(mem_) >= sizeof(detail::standalone_target<Ret, Args...>), "");
      new (&mem_) detail::standalone_target<Ret, Args...>(func);
   }

   template <typename T>
   fun_ptr(Ret (T::*method)(Args...), T* object)
      : mem_type_id_{detail::IdGen<
           detail::mutable_member_target<Ret, T, Args...>>::get_unique_id()}
   {
      static_assert(
         alignof(decltype(mem_)) %
               alignof(detail::mutable_member_target<Ret, T, Args...>) ==
            0,
         "");
      static_assert(sizeof(mem_) >=
                       sizeof(detail::mutable_member_target<Ret, T, Args...>),
                    "");
      new (&mem_)
         detail::mutable_member_target<Ret, T, Args...>(method, object);
   }

   template <typename T>
   fun_ptr(Ret (T::*method)(Args...) const, const T* object)
      : mem_type_id_{detail::IdGen<
           detail::const_member_target<Ret, T, Args...>>::get_unique_id()}
   {
      static_assert(
         alignof(decltype(mem_)) %
               alignof(detail::const_member_target<Ret, T, Args...>) ==
            0,
         "");
      static_assert(sizeof(mem_) >=
                       sizeof(detail::const_member_target<Ret, T, Args...>),
                    "");
      new (&mem_) detail::const_member_target<Ret, T, Args...>(method, object);
   }

   template <typename T>
   fun_ptr(T f)
      : mem_type_id_{detail::IdGen<
           detail::function_object_target<Ret, T, Args...>>::get_unique_id()}
   {
      static_assert(
         alignof(decltype(mem_)) %
               alignof(detail::function_object_target<Ret, T, Args...>) ==
            0,
         "");
      static_assert(sizeof(mem_) >=
                       sizeof(detail::function_object_target<Ret, T, Args...>),
                    "");
      new (&mem_) detail::function_object_target<Ret, T, Args...>(f);
   }

   fun_ptr(const fun_ptr& rhs) noexcept : mem_type_id_{rhs.mem_type_id_}
   {
      rhs.get_target().clone(&mem_);
   }

   ~fun_ptr() noexcept
   {
      get_target().~target_base();
   }

   fun_ptr& operator=(std::nullptr_t) noexcept
   {
      get_target().~target_base();
      new (&mem_) detail::null_target<Ret, Args...>();
      mem_type_id_ =
         detail::IdGen<detail::null_target<Ret, Args...>>::get_unique_id();
      return *this;
   }

   fun_ptr& operator=(const fun_ptr& rhs) noexcept
   {
      get_target().~target_base();
      rhs.get_target().clone(&mem_);
      mem_type_id_ = rhs.mem_type_id_;
      return *this;
   }

   explicit operator bool() const noexcept
   {
      return mem_type_id_ !=
             detail::IdGen<detail::null_target<Ret, Args...>>::get_unique_id();
   }

   template <typename... FwArgs>
   Ret operator()(FwArgs&&... args) const
   {
      return get_target().invoke(std::forward<FwArgs>(args)...);
   }

private:
   detail::target_base<Ret, Args...>& get_target() noexcept
   {
      return *reinterpret_cast<detail::target_base<Ret, Args...>*>(&mem_);
   }

   const detail::target_base<Ret, Args...>& get_target() const noexcept
   {
      return *reinterpret_cast<const detail::target_base<Ret, Args...>*>(&mem_);
   }

   typename std::aligned_union<
      1, detail::standalone_target<Ret, Args...>,
      detail::mutable_member_target<Ret, detail::SomeClass, Args...>,
      detail::const_member_target<Ret, detail::SomeClass, Args...>,
      detail::function_object_target<Ret, detail::SomeClass, Args...>>::type
      mem_;

   // uniquely identifies the target derived type stored in mem_ (optimization
   // used to remove use of RTTI)
   detail::unique_id_t mem_type_id_;

   template <typename FrRet, typename... FrArgs>
   friend bool operator==(const fun_ptr<FrRet(FrArgs...)>& lhs,
                          const fun_ptr<FrRet(FrArgs...)>& rhs) noexcept;
};

template <typename Ret, typename... Args>
inline bool operator==(const fun_ptr<Ret(Args...)>& lhs,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept
{
   return (lhs.mem_type_id_ == rhs.mem_type_id_) &&
          lhs.get_target().check_equality(rhs.get_target());
}

template <typename Ret, typename... Args>
inline bool operator!=(const fun_ptr<Ret(Args...)>& lhs,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept
{
   return !(lhs == rhs);
}

template <typename Ret, typename... Args>
inline bool operator==(const fun_ptr<Ret(Args...)>& lhs,
                       std::nullptr_t) noexcept
{
   return !static_cast<bool>(lhs);
}

template <typename Ret, typename... Args>
inline bool operator!=(const fun_ptr<Ret(Args...)>& lhs,
                       std::nullptr_t) noexcept
{
   return !(lhs == nullptr);
}

template <typename Ret, typename... Args>
inline bool operator!=(std::nullptr_t,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept
{
   return !(rhs == nullptr);
}

template <typename Ret, typename... Args>
inline bool operator==(std::nullptr_t,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept
{
   return rhs == nullptr;
}

template <typename Ret, typename... Args>
inline fun_ptr<Ret(Args...)> make_fun(Ret (*fp)(Args...))
{
   return fun_ptr<Ret(Args...)>(fp);
}

template <typename Ret, typename T, typename... Args>
inline fun_ptr<Ret(Args...)> make_fun(Ret (T::*fp)(Args...), T* obj)
{
   return fun_ptr<Ret(Args...)>(fp, obj);
}

template <typename Ret, typename T, typename... Args>
inline fun_ptr<Ret(Args...)> make_fun(Ret (T::*fp)(Args...) const, const T* obj)
{
   return fun_ptr<Ret(Args...)>(fp, obj);
}

template <typename T>
inline auto make_fun(T lambda)
   -> decltype(make_fun(&T::operator(), static_cast<T*>(nullptr)))
{
   return decltype(make_fun(&T::operator(), static_cast<T*>(nullptr)))(lambda);
}

} // namespace util3

#endif // UTIL_FUN_PTR3_H
