#ifndef FUN_PTR_INTERFACE_H
#define FUN_PTR_INTERFACE_H

// fun_ptr is a value type with its only salient attribute being its target.
// The target is the function (standalone, mutable  member, const member,
// function object/lambda, or nullptr) and object instance when applicable.
// Unlike std::function (or boost::function) class fun_ptr is designed to have
// function pointer like semantics. fun_ptr supports equality operations which
// std::function does not. The target is invoked via operator(). Invoking a
// fun_ptr with a nullptr target is not permitted. operator= changes the target
// of the fun_ptr and is the only salient mutating operation that can be
// performed on a fun_ptr.
//
// Equality comparison. fun_ptrs with targets that point to the same object
// instance and function compare equal. Two fun_ptrs with different target
// types (i.e mutable member function and const member function) will compare
// unequal. Two fun_ptrs with different object instances will compare unequal.
// The equality comparison result is unspecified if the target type and object
// instances are the same but one or both of the member functions is virtual
// (because C++ spec does not specify the result of comparison of virtual
// funtion pointers).
//
// Function object(or lambda) types are by convention (see Meyers) passed by
// value (and often are temporary objects). When a fun_ptr is constructed with
// a function object it constructs a copy of the function object and then
// manages its lifetime. The function object will live so long as any fun_ptr
// references it. As a result two fun_ptrs constructed from the same function
// object will not compare equal (see example 1) because the targets are
// different instances of the function object type. However, copying a fun_ptr
// does not construct a new function object. The newly created fun_ptr will
// share the same target and extend the life of the function object should the
// original fun_ptr go out of scope.
//
// For member function targets it is up to the user to ensure the target
// references a valid object. That is invoking a fun_ptr target on an object
// that no longer exists is not permitted.
//
// fun_ptr has shallow const and shallow copy semantics. Target function object
// lifetimes are managed by the fun_ptr class - not necessarily by a single
// fun_ptr instance but possibly by any/all the fun_ptr instances - it is a
// shared target. The shared function object does not participate in the
// constness of fun_ptr.
//
// The question might be asked: is fun_ptr thread safe? Recall that fun_ptr
// was designed to have function pointer like semantics so the answer is the
// same as the answer to the question: "is a function pointer thread safe?". It
// depends on the implementation of the function its points to. For standalone
// functions the function would need to be reentrant. For a member function or
// function object with state it would need to be internally synchronized.
//
// Examples 1: function object equality
// auto lam = []() { /*...*/ };
// auto d1 = make_fun(lam);
// auto d2 = d1; //copy constructor
// assert(d2 == d1); //always true alfter a copy constructor
// d1(); //invoke operator() const
// assert(d2 == d1); //since operator() is const equality was not changed
// d2 = make_fun(lam);
// assert(d1 != d2); //a new instance of the lambda was created so d1 != d2
// d1 = d2; //copy assignment
// assert(d1 == d2); //always true alfter a copy assignment
//
// Examples 2: member function object equality
// auto d1 = make_fun(&Some_class::some_mem_fun, &some_class);
// auto d2 = d1; //copy constructor
// assert(d1 == d2); //always true alfter a copy constructor
// d1(); //invoke operator() const
// assert(d2 == d1); //since operator() is const equality was not changed
// d2 = make_fun(&Some_class::some_mem_fun, &some_class);
// assert(d1 == d2); //target is now the same so equal (note difference /w ex.1)
// d1 = d2; //copy assignment
// assert(d1 == d2); //always true alfter a copy assignment
//
// Examples 3: implications of equality. The result of rand() is don't care. The
//             ovservable behavior is identical - lam.operator() is called 1000
//             times on the same instance of lam;
// auto d1 = make_fun([]() { /*...*/ });
// auto d2 = d1; //copy constructor - shares target with d1.
// for (int i = 0; i < 1000; ++i) {
//    if (rand() % 2 == 0) {
//       d1();
//    } else {
//       d2();
//    }
//    assert(d1 == d2);
// }

template <typename FuncSignature>
class fun_ptr;

template <typename Ret, typename... Args>
class fun_ptr<Ret(Args...)>
{
public:
   fun_ptr() noexcept;

   fun_ptr(std::nullptr_t) noexcept;

   // not declared noexcept because of a narrow contract and optionally dynamic
   // memory allocation is allowed.
   fun_ptr(Ret (*func)(Args...));

   // not declared noexcept because of a narrow contract and optionally dynamic
   // memory allocation is allowed.
   template <typename T>
   fun_ptr(Ret (T::*method)(Args...), T* object);

   // not declared noexcept because of a narrow contract and optionally dynamic
   // memory allocation is allowed.
   template <typename T>
   fun_ptr(Ret (T::*method)(Args...) const, const T* object);

   // not declared noexcept because optionally dynamic memory allocation is
   // allowed.
   template <typename T>
   fun_ptr(T f);

   // noexcept??? Sharing a target does not require dynamic memory allocation.
   // But do I want to preclude its use for future implementations?
   fun_ptr(const fun_ptr& rhs) noexcept;

   fun_ptr(fun_ptr&& rhs) noexcept;

   ~fun_ptr() noexcept;

   fun_ptr& operator=(std::nullptr_t) noexcept;

   // noexcept??? Sharing a target does not require dynamic memory allocation.
   // But do I want to preclude its use for future implementations?
   fun_ptr& operator=(const fun_ptr& rhs) noexcept;

   fun_ptr& operator=(fun_ptr&& rhs) noexcept;

   explicit operator bool() const noexcept;

   template <typename... FwArgs>
   Ret operator()(FwArgs... args) const;
};

template <typename Ret, typename... Args>
inline bool operator==(const fun_ptr<Ret(Args...)>& lhs,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept;

template <typename Ret, typename... Args>
inline bool operator!=(const fun_ptr<Ret(Args...)>& lhs,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept;

template <typename Ret, typename... Args>
inline bool operator==(const fun_ptr<Ret(Args...)>& lhs,
                       std::nullptr_t) noexcept;

template <typename Ret, typename... Args>
inline bool operator!=(const fun_ptr<Ret(Args...)>& lhs,
                       std::nullptr_t) noexcept;

template <typename Ret, typename... Args>
inline bool operator==(std::nullptr_t,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept;

template <typename Ret, typename... Args>
inline bool operator!=(std::nullptr_t,
                       const fun_ptr<Ret(Args...)>& rhs) noexcept;

template <typename Ret, typename... Args>
inline fun_ptr<Ret(Args...)> make_fun(Ret (*fp)(Args...));

template <typename Ret, typename T, typename... Args>
inline fun_ptr<Ret(Args...)> make_fun(Ret (T::*fp)(Args...), T* obj);

template <typename Ret, typename T, typename... Args>
inline fun_ptr<Ret(Args...)> make_fun(Ret (T::*fp)(Args...) const,
                                      const T* obj);

template <typename T>
inline auto make_fun(T functor)
   -> decltype(make_fun(&T::operator(), (T*) nullptr));

#endif // FUN_PTR_INTERFACE_H
